#!/usr/bin/env python
#
# Copyright (c) 2024, Ryan Galloway (ryan@rsgalloway.com)
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#  - Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
#  - Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
#  - Neither the name of the software nor the names of its contributors
#    may be used to endorse or promote products derived from this software
#    without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#

__doc__ = """
Contains logging functions and classes.
"""

import logging

from distman.config import LOG_LEVEL

log = logging.Logger("distman")
log.setLevel(LOG_LEVEL)
log.addHandler(logging.NullHandler())


def setup_stream_handler():
    """Adds a new stdout stream handler."""
    for h in log.handlers:
        if h.name == log.name and "StreamHandler" in str(h):
            del log.handlers[log.handlers.index(h)]
    stream_hanlder = logging.StreamHandler()
    stream_hanlder.set_name(log.name)
    stream_hanlder.setFormatter(logging.Formatter("%(message)s"))
    log.addHandler(stream_hanlder)



#!/usr/bin/env python
#
# Copyright (c) 2024, Ryan Galloway (ryan@rsgalloway.com)
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#  - Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
#  - Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
#  - Neither the name of the software nor the names of its contributors
#    may be used to endorse or promote products derived from this software
#    without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#

__doc__ = """
Contains source file distribution classes and functions.
"""

import json
import os

import git

from distman import config, util
from distman.logger import log


class Source(object):
    """File source base class."""

    def __init__(self):
        super(Source, self).__init__()
        self.author = os.getenv("USERNAME", os.getenv("USER", "unknown"))
        self.changed_files = []
        self.directory = "."
        self.name = ""
        self.path = ""
        self.root = None

    def get_targets(self):
        """Returns the list of targets in the dist file."""
        if self.root:
            return self.root.get(config.TAG_TARGETS)
        return None

    def read_dist_file(self, directory="."):
        """Opens and parses the dist file.

        :param directory: Path to directory containing the dist file.
        :return: True if successful.
        """
        if not os.path.isdir(directory):
            log.info("%s is not a directory", directory)
            return False

        dist_file = os.path.join(directory, config.DIST_FILE)
        if not os.path.exists(dist_file):
            log.info("%s does not exist", dist_file)
            return False

        self.directory = directory

        try:
            with open(dist_file, "r") as jsonFile:
                json_data = json.load(jsonFile)
        except Exception as e:
            log.error("Failed to parse dist file: %s", str(e))
            return False

        self.root = json_data
        self.author = self.root.get(config.TAG_AUTHOR, util.get_user())

        log.info("Author: %s" % self.author)
        if config.TAG_VERSION in self.root:
            self.dist_file_version = self.root[config.TAG_VERSION]
            if int(self.dist_file_version) < config.DIST_FILE_VERSION:
                log.warning(
                    "WARNING: Old dist file version: %s (current %d)"
                    % (self.dist_file_version, config.DIST_FILE_VERSION)
                )
            elif int(self.dist_file_version) > config.DIST_FILE_VERSION:
                log.error(
                    "ERROR: This dist file is newer than this script version: %s "
                    "(currrent %d)" % (self.dist_file_version, config.DIST_FILE_VERSION)
                )
                self.root = None
                return False

        return True


class GitRepo(Source):
    """Git repo file source base class."""

    def __init__(self):
        super(GitRepo, self).__init__()
        self.branch_name = ""
        self.head = ""
        self.repo = None
        self.short_head = ""

    def get_repo_files(self, start=None):
        """Generator that yields relative file paths tracked by this git repo.

        :param start: Starting directory.
        :return: List of relative file paths.
        """
        from pathlib import Path

        repo_root = Path(self.repo.working_tree_dir).resolve()

        # resolve the start directory relative to the repo root
        start_path = (repo_root / start).resolve()
        if not start_path.is_dir():
            raise ValueError(
                f"Start directory '{start}' does not exist or is not a directory."
            )

        # get the list of tracked files and filter by start_dir
        tracked_files = []
        for item in self.repo.tree().traverse():
            item_path = repo_root / item.path
            if item_path.is_file() and start_path in item_path.parents:
                # append relative path from the repo root
                tracked_files.append(str(item_path.relative_to(repo_root)))

        return tracked_files

    def get_path(self):
        """Get the git repo path for the dist info file."""
        if self.repo and self.repo.remotes:
            if "origin" in [remote.name for remote in self.repo.remotes]:
                return self.repo.remotes.origin.url
            else:
                return self.repo.remotes[0].url
        return self.path

    def read_git_info(self):
        """Read git repo information."""
        self.branch_name = ""
        self.head = ""
        self.short_head = ""
        self.path = os.getcwd().replace("\\", "/")
        self.name = os.path.basename(self.path)
        self.changed_files = []

        try:
            self.repo = git.Repo(self.directory)
            self.head = self.repo.head.commit.hexsha
            self.short_head = self.head[: config.LEN_HASH]
            self.path = self.get_path()

            # remove user name and protocol from path
            if "@" in self.path:
                self.path = self.path[self.path.find("@") + 1 :]

            # change ssh style path to look like https style
            # self.path = self.path.replace(":", "/")
            self.name = self.path[self.path.rfind("/") + 1 :]
            if self.name.endswith(".git"):
                self.name = self.name[:-4]

            # this will generate TypeError if detached HEAD
            branch = self.repo.active_branch
            self.branch_name = branch.name
            if branch.name not in config.MAIN_BRANCHES:
                log.warning("WARNING: Not on master branch")

        except git.InvalidGitRepositoryError:
            log.warning("Not in a git repository directory")

        except (AttributeError, TypeError) as e:
            log.warning("WARNING: %s", str(e))

        except Exception as e:
            log.warning("ERROR: error reading git repo\n%s", str(e))

        log.info("Name: %s" % self.name)
        log.info("Path: %s" % self.path)
        if self.branch_name:
            log.info("Branch: %s" % self.branch_name)
            log.info("Head: %s (%s)" % (self.head, self.short_head))

        return True

    def is_git_behind(self):
        """Checks for upstream commits on the repo."""
        if not self.branch_name:
            return False

        if not self.repo.remotes:
            return False

        try:
            upstream_commits = list(
                self.repo.iter_commits(f"origin/{self.branch_name}..{self.branch_name}")
            )
            if upstream_commits:
                log.error(
                    "Directory is %d commits behind remote repository. "
                    "Run: git pull from origin first or use --force."
                    % len(upstream_commits)
                )
                return True

        except Exception as err:
            log.error("Error checking remote branch: %s", str(err))
            return True

        return False

    def git_changed_files(self):
        """Returns list of changed files (in staging or untracked).

        :return: List of changed files.
        """
        if not self.repo:
            return []

        # get list of changed files
        changed_files = [
            os.path.abspath(self.directory + os.path.sep + item.a_path)
            for item in self.repo.index.diff(None)
        ]

        # get list of staged files
        changed_files += [
            os.path.abspath(self.directory + os.path.sep + item.a_path)
            for item in self.repo.index.diff("HEAD")
        ]

        # get list of untracked files
        changed_files += [
            os.path.abspath(self.directory + os.path.sep + item)
            for item in self.repo.untracked_files
        ]

        return changed_files



#!/usr/bin/env python
#
# Copyright (c) 2024, Ryan Galloway (ryan@rsgalloway.com)
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#  - Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
#  - Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
#  - Neither the name of the software nor the names of its contributors
#    may be used to endorse or promote products derived from this software
#    without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#

__doc__ = """
Command line interface for distman: simple software distribution system.

Usage:

    $ distman [LOCATION] [OPTIONS]
"""

import argparse
import os
import sys

from distman import Distributor, config, util
from distman.logger import setup_stream_handler

setup_stream_handler()


def parse_args():
    """Parse command line arguments."""
    from distman import __version__

    parser = argparse.ArgumentParser(
        description=__doc__, formatter_class=argparse.RawTextHelpFormatter
    )
    parser.add_argument(
        "location",
        metavar="LOCATION",
        nargs="?",
        default=".",
        help="The directory containing dist file. Default is current directory",
    )
    parser.add_argument(
        "-t",
        "--target",
        metavar="TARGET",
        help="The source TARGET in the dist file (supports wildcards)",
    )
    parser.add_argument(
        "-s",
        "--show",
        action="store_true",
        help="Show current distributed versions only",
    )
    parser.add_argument(
        "-c",
        "--commit",
        metavar="HASH",
        help="Change TARGET version number to point to commit HASH",
    )
    parser.add_argument(
        "-n",
        "--number",
        metavar="NUMBER",
        help="Change TARGET version number to point to NUMBER",
    )
    parser.add_argument(
        "--reset",
        action="store_true",
        help="Reset version for TARGET to point to latest version",
    )
    parser.add_argument(
        "--delete",
        action="store_true",
        help="Delete dist TARGET from deployment folder.",
    )
    parser.add_argument(
        "-y",
        "--yes",
        action="store_true",
        help="Answer yes to all questions, skipping user interaction",
    )
    parser.add_argument(
        "-f",
        "--force",
        action="store_true",
        help="Force action, e.g. disting uncommitted file changes",
    )
    parser.add_argument(
        "-d",
        "--dryrun",
        action="store_true",
        help="Do a dry run, no actions will be performed",
    )
    parser.add_argument(
        "--verbose",
        action="store_true",
        help="Show verbose information",
    )
    parser.add_argument(
        "-v",
        "--version",
        action="version",
        version=f"distman {__version__}",
    )

    args = parser.parse_args()
    return args


def main():
    """Main thread."""

    args = parse_args()

    if not os.path.isdir(args.location):
        print("%s is not a directory" % args.location)
        return 1

    if sum([bool(args.commit), bool(args.number), bool(args.reset)]) > 1:
        print("--commit,--number and --reset are mutually exclusive")
        return 1

    distributor = Distributor()

    # process the requested location
    if not distributor.read_dist_file(args.location):
        return 1

    # remove target(s), symlink privs not needed
    if args.delete:
        if distributor.delete_target(args.target, yes=args.yes, dryrun=args.dryrun):
            return 0
        else:
            return 1

    # on windows make sure SeCreateSymbolicLinkPrivilege privilege is held
    # (must be done after Distributor object is created)
    if os.name == "nt":
        if not util.check_symlinks():
            return 1

    if args.number or args.commit or args.reset:
        if args.reset:
            if distributor.reset_file_version(
                args.target,
                dryrun=args.dryrun,
            ):
                return 0
            else:
                return 1

        elif args.number:
            if not args.target:
                print("No target specified to change version")
                return 2
            else:
                target_file = args.target
                target_version = args.number
            try:
                target_version = int(target_version)
            except Exception:
                print("Invalid version number: %s" % args.number[1])
                return 2
            target_commit = ""

        else:
            if not args.target:
                print("No target specified to change version")
                return 2
            else:
                target_file = args.target
                target_commit = args.commit
            if len(target_commit) < config.LEN_MINHASH:
                print("Hashes must be at least %d characters" % config.LEN_MINHASH)
                return 0
            target_version = 0

        # file version change
        if distributor.change_file_version(
            target_file, target_commit, target_version, dryrun=args.dryrun
        ):
            return 0
        else:
            return 1

    # do file distribution
    if distributor.dist(
        target=args.target,
        show=args.show,
        force=args.force,
        yes=args.yes,
        dryrun=args.dryrun,
        verbose=args.verbose,
    ):
        return 0

    return 1


if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/env python
#
# Copyright (c) 2024, Ryan Galloway (ryan@rsgalloway.com)
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#  - Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
#  - Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
#  - Neither the name of the software nor the names of its contributors
#    may be used to endorse or promote products derived from this software
#    without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#

__doc__ = """
Contains tests for the dist module.
"""

import json
import unittest
from unittest.mock import patch

from distman import Distributor


class TestDistributor(unittest.TestCase):
    """Tests for the Distributor class."""

    def setUp(self):
        self.distributor = Distributor()

    def test_read_dist_file(self):
        # Mock the dist file content
        dist_file_content = {
            "author": "John Doe",
            "targets": {
                "target1": {
                    "sourcePath": "path/to/source1",
                    "destPath": "path/to/dest1",
                },
                "target2": {
                    "sourcePath": "path/to/source2",
                    "destPath": "path/to/dest2",
                },
            },
        }

        # Mock the open() function to return the dist file content
        with patch("builtins.open", create=True) as mock_open:
            mock_open.return_value.__enter__.return_value.read.return_value = (
                json.dumps(dist_file_content)
            )

            # Call the method to read the dist file
            result = self.distributor.read_dist_file()

            # Assert that the method returns True
            self.assertTrue(result)

            # Assert that the distributor's root attribute is set correctly
            self.assertEqual(self.distributor.root, dist_file_content)

    def test_get_targets(self):
        # Set the distributor's root attribute
        self.distributor.root = {
            "targets": {
                "target1": {
                    "sourcePath": "path/to/source1",
                    "destPath": "path/to/dest1",
                },
                "target2": {
                    "sourcePath": "path/to/source2",
                    "destPath": "path/to/dest2",
                },
            }
        }

        # Call the method to get the targets
        targets = self.distributor.get_targets()

        # Assert that the targets are returned correctly
        self.assertEqual(targets, self.distributor.root["targets"])

    def test_get_files(self):
        # Set the distributor's directory attribute
        self.distributor.directory = "/path/to/directory"

        # Mock the util.walk() function to return a list of files
        with patch("distman.util.walk") as mock_walk:
            mock_walk.return_value = [
                "/path/to/directory/file1.txt",
                "/path/to/directory/file2.txt",
                "/path/to/directory/subdirectory/file3.txt",
            ]

            # Call the method to get the files
            files = self.distributor.get_files("/path/to/directory")

            # Assert that the files are returned correctly
            self.assertEqual(
                files,
                [
                    "/path/to/directory/file1.txt",
                    "/path/to/directory/file2.txt",
                    "/path/to/directory/subdirectory/file3.txt",
                ],
            )


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
#
# Copyright (c) 2024, Ryan Galloway (ryan@rsgalloway.com)
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#  - Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
#  - Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
#  - Neither the name of the software nor the names of its contributors
#    may be used to endorse or promote products derived from this software
#    without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#

__doc__ = """
Contains utility functions and classes.
"""

import ctypes
import fnmatch
import os
import re
import shutil

from distman import config
from distman.logger import log

# cache regex pattern that matches any patterns in IGNORABLE
IGNORABLE_PATHS = re.compile(
    "(" + ")|(".join([fnmatch.translate(i) for i in config.IGNORABLE]) + ")"
)


def add_symlink_support():
    """Adds symlink support for Windows."""

    def symlink_ms(source, link_name, target_is_directory=False):
        csl = ctypes.windll.kernel32.CreateSymbolicLinkW
        csl.argtypes = (ctypes.c_wchar_p, ctypes.c_wchar_p, ctypes.c_uint32)
        csl.restype = ctypes.c_ubyte
        flags = 1 if target_is_directory else 0
        if csl(link_name, source.replace("/", "\\"), flags) == 0:
            raise ctypes.WinError()

    os.symlink = symlink_ms


def check_symlinks():
    """Checks if it is possible to create symbolic links by creatiung a temp file
    and trying to create a link to it.

    :returns: True if symbolic links can be created.
    """
    import tempfile

    temp_file = tempfile.mktemp()
    link_file = tempfile.mktemp()

    with open(temp_file, "a"):
        pass

    try:
        os.symlink(temp_file, link_file)

    except OSError:
        log.warning("Privileges to create symbolic links are required.")
        log.warning(
            "Run as Administrator or change system settings for "
            "SeCreateSymbolicLinkPrivilege."
        )
        os.remove(temp_file)
        return False

    os.remove(link_file)
    os.remove(temp_file)

    return True


def get_user():
    """Returns the current user name."""
    return os.getenv("USER", os.getenv("USERNAME", "unknown"))


def has_hidden_attr(filepath):
    """Checks if file has hidden file attribute (windows only).

    :param filepath: file system path.
    :returns: True if file is hidden.
    """

    try:
        attrs = ctypes.windll.kernel32.GetFileAttributesW(str(filepath))
        assert attrs != -1
        result = bool(attrs & 2)

    except (AttributeError, AssertionError):
        result = False

    return result


def is_file_hidden(filepath):
    """Cross platform check if file is hidden. Checks if file name begins
    with period or has hidden attribute.

    :param filepath: file system path.
    :returns: True if file is hidden.
    """
    name = os.path.basename(os.path.abspath(filepath))
    return name.startswith(".") or has_hidden_attr(filepath)


def is_ignorable(filepath):
    """Returns True if path is ignorable. Checks path against patterns
    in the ignorables list, as well as dot files.

    :param path: a file system path.
    :returns: True if filepath is ignorable.
    """

    if is_file_hidden(filepath):
        return True

    return re.search(IGNORABLE_PATHS, filepath) is not None


def normalize_path(path):
    """Normalizes paths by changing separators to forward slashes and removing
    trailing slashes.

    :param path: file system path.
    :returns: normalized path.
    """

    if not path:
        return path

    path = path.replace("\\", "/")

    if path[-1] == "/":
        path = path[:-1]

    return path


def get_dist_info(dest, ext=config.DIST_INFO_EXT):
    """Returns the dist info file path, e.g.

        /path/to/desploy/prod/.foobar.py.dist

    The dist info files are hidden dot files used to store distribution
    information and tell distman if a file has been previously disted.

    :param dest: destination directory.
    :param ext: file extension.
    """
    folder, original_name = os.path.split(dest)
    return os.path.join(folder, f".{original_name}{ext}")


def write_dist_info(dest, dist_info):
    """Writes distribution information to a file.

    :param dest: Path to destination directory.
    :param dist_info: Dictionary of distribution information.
    """
    distinfo = get_dist_info(dest=dest)
    log.debug("Writing dist info to %s" % distinfo)
    with open(distinfo, "w") as outFile:
        for key, value in dist_info.items():
            outFile.write(f"{key}: {value}\n")


def create_dest_folder(dest, dryrun=False, yes=False):
    """Creates destination folder if it does not exist.

    :param dest: destination file path.
    :param dryrun: dry run flag.
    :param yes: yes flag.
    :returns: True if destination folder was created.
    """
    dest_dir = os.path.dirname(dest)

    if not os.path.exists(dest_dir):
        log.info("Creating destination directory '%s'" % dest_dir)
        if not dryrun:
            try:
                os.makedirs(dest_dir)
            except Exception as e:
                log.info(
                    "ERROR: Failed to create directory '%s': %s" % (dest_dir, str(e))
                )
                return False
    elif not os.path.isdir(dest_dir):
        log.info("Directory not found: %s" % dest_dir)
        return False

    # if dist info file does not exist means this is a new target
    distinfo = get_dist_info(dest)
    if not os.path.exists(distinfo):
        if os.path.exists(dest):
            question = (
                "Target '%s' already exists as a %s and will "
                "be deleted, continue?"
                % (dest, "dir" if os.path.isdir(dest) else "file")
            )
            if not yes and not yesNo(question):
                return False
        log.info("Initializing: %s" % dest)


def full_path(start, relative_path):
    """Returns the full path from a relative path.

    :param start: starting directory.
    :param relative_path: relative path.
    :returns: full path.
    """
    if not relative_path:
        raise Exception("Empty path")

    if (
        relative_path[0] == "/"
        or relative_path[0] == "\\"
        or (
            len(relative_path) > 1
            and (relative_path[1] == ":" or relative_path.startswith(".."))
        )
    ):
        path = os.path.abspath(relative_path)
        temp_relative_path = os.path.relpath(path, start=start)
        if temp_relative_path.startswith(".."):
            raise Exception("Path below start directory")
        return path

    if not start:
        start = os.getcwd()

    return os.path.abspath(start + os.path.sep + relative_path)


def remove_object(path, recurse=False):
    """Deletes a file or directory tree.

    :param path: file system path.
    :param recurse: recursively delete directory tree.
    """
    try:
        if os.path.isdir(path):
            if recurse:
                shutil.rmtree(path)
            else:
                os.rmdir(path)
        else:
            os.remove(path)

    except OSError:
        try:
            # no way to tell if a directory is a symlink:
            # os.path.isdir() will return True
            # try to delete as file if fails
            os.remove(path)
        except OSError as e:
            log.error("Error removing '%s': %s" % (path, str(e)))


def yesNo(question):
    """Displays question text to user and reads yes/no input.

    :param question: question text.
    :returns: True if user answers yes.
    """
    while True:
        answer = input(question + " (y/n): ").lower().strip()
        if answer in ("y", "yes", "n", "no"):
            return answer in ("y", "yes")
        else:
            print("You must answer yes or no.")


def walk(path, exclude_ignorables=True):
    """Generator that yields relative file paths that are not ignorable.

    :param path: file system path.
    :param exclude_ignorables: exclude ignorable files.
    :returns: generator of file paths.
    """
    if not is_ignorable(path) and os.path.isfile(path):
        yield path
    for dirname, dirs, files in os.walk(path, topdown=True):
        if exclude_ignorables and is_ignorable(dirname):
            continue
        for d in dirs:
            if exclude_ignorables and is_ignorable(d):
                dirs.remove(d)
        for name in files:
            if not is_ignorable(name):
                yield os.path.join(dirname, name)



#!/usr/bin/env python
#
# Copyright (c) 2024, Ryan Galloway (ryan@rsgalloway.com)
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#  - Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
#  - Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
#  - Neither the name of the software nor the names of its contributors
#    may be used to endorse or promote products derived from this software
#    without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#

__doc__ = """
Contains default config and settings.
"""

import os
import platform

PLATFORM = platform.system().lower()

# default environment settings
DEFAULT_ENV = {
    "ENV": "prod",
    "HOME": os.getenv("HOME"),
    "ROOT": {
        "darwin": "{HOME}/Library/Application Support/pipe",
        "linux": "{HOME}/.local/pipe",
        "windows": "C:\\ProgramData\\pipe",
    }.get(PLATFORM, "./pipe/{ENV}"),
    "DEPLOY_ROOT": "{ROOT}/{ENV}",
}

# dist file settings
DIST_FILE = "dist.json"
DIST_FILE_VERSION = 1
DIST_INFO_EXT = ".dist"
DIR_VERSIONS = "versions"

# logging settings
LOG_LEVEL = os.getenv("LOG_LEVEL", "INFO")
DRYRUN_MESSAGE = "NOTICE: Dry run (no changes will be made)"

# ignorable files and directories
IGNORABLE = [
    "*~",
    ".git*",
    ".env",
    ".venv",
    "*.bup",
    "*.swp",
    "*.temp*",
    "*.tmp",
    "venv*",
    "MANIFEST*",
    "__pycache__",
    "Thumbs.db",
    ".DS_Store",
]

# git repo settings
LEN_HASH = 7
LEN_MINHASH = 4
MAIN_BRANCHES = ["master", "main"]

# embedded path tokens
PATH_TOKEN_OPEN = "{"
PATH_TOKEN_CLOSE = "}"

# dist file keys/tags
TAG_AUTHOR = "author"
TAG_DESTPATH = "destination"
TAG_SOURCEPATH = "source"
TAG_VERSION = "version"
TAG_TARGETS = "targets"



#!/usr/bin/env python
#
# Copyright (c) 2024, Ryan Galloway (ryan@rsgalloway.com)
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#  - Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
#  - Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
#  - Neither the name of the software nor the names of its contributors
#    may be used to endorse or promote products derived from this software
#    without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#

__doc__ = """
Contains file distribution classes and functions.
"""

import filecmp
import fnmatch
import os
import shutil
import time

from distman import config, util
from distman.logger import log
from distman.source import GitRepo


class Distributor(GitRepo):
    """File distribution class."""

    def __init__(self):
        super(Distributor, self).__init__()
        self.__add_symlink_support()

    @staticmethod
    def __add_symlink_support():
        """Adds symlink support for platforms that lack it."""
        os_symlink = getattr(os, "symlink", None)
        if not callable(os_symlink):
            util.add_symlink_support()

    @staticmethod
    def __get_file_versions(target):
        """Find the highest numeric version number for a file.

        :param target: Path to file to check.
        :return: List of tuples with version number and file.
        """
        filedir = os.path.dirname(target) + "/" + config.DIR_VERSIONS
        if not os.path.exists(filedir):
            return []

        filename = os.path.basename(target)
        version_list = []

        for f in os.listdir(filedir):
            file_name_length = len(filename)
            # get files that match <target>.<version>.<commit>
            if (
                f.startswith(filename)
                and len(f) > file_name_length + 1
                and f[file_name_length] == "."
                and str(f[file_name_length + 1]).isnumeric()
            ):
                # parse the number from the rest of the file name
                info = f[file_name_length + 1 :]
                dotPos = info.find(".")
                if -1 != dotPos:
                    ver = int(info[:dotPos])
                else:
                    ver = int(info)
                commit = ""
                if -1 != dotPos:
                    # trim potential remaining dotted portions
                    dotPos2 = info.find(".", dotPos + 1)
                    if -1 == dotPos2:
                        commit = info[dotPos + 1 :]
                    else:
                        commit = info[dotPos + 1 : dotPos2]
                    # trim '-forced' if present
                    dashPos = commit.find("-")
                    if -1 != dashPos:
                        commit = commit[:dashPos]
                version_list.append((filedir + "/" + f, ver, commit))

        return sorted(version_list, key=lambda tup: tup[1])

    @staticmethod
    def __hashes_equal(commitA, commitB):
        """Compares two hash strings regardless of length or case

        :param commitA: First hash string.
        :param commitB: Second hash string.
        :return: True if hashes are equal.
        """
        if len(commitA) > len(commitB):
            return commitA.upper().startswith(commitB.upper())
        else:
            return commitB.upper().startswith(commitA.upper())

    def __copy_file(self, source, dest):
        """Copies a file, and converts line endings to linux LF, preserving
        original source file mode.

        :param source: Path to source file.
        :param dest: Path to destination file.
        """
        try:
            destdir = os.path.dirname(dest)
            if not os.path.isdir(destdir):
                os.makedirs(destdir)
            with open(source, "r") as infile, open(dest, "wb") as outfile:
                for line in infile:
                    text = line.rstrip("\r\n")
                    outfile.write((text + "\n").encode("UTF-8"))
        except UnicodeDecodeError:
            shutil.copy2(source, dest)
        except Exception as e:
            log.error("File copy error: %s" % str(e))
        finally:
            mode = os.stat(source).st_mode
            os.chmod(dest, mode)

    def __copy_directory(self, source, dest):
        """Recursively copies a directory (ignores hidden files).

        :param source: Path to source directory.
        :param dest: Path to destination directory.
        """
        source = os.path.relpath(source)
        all_files = self.get_files(source)

        for filepath in all_files:
            if source == ".":
                target = os.path.join(dest, filepath)
            else:
                target = os.path.join(dest, filepath[len(source) + 1 :])
            self.__copy_file(filepath, target)

    def __copy_object(self, source, dest):
        """Copies a file or directory recursively (ignores hidden files).

        :param source: Path to source file or directory.
        :param dest: Path to destination file or directory.
        """
        if os.path.isfile(source):
            self.__copy_file(source, dest)
        elif os.path.isdir(source):
            self.__copy_directory(source, dest)
        else:
            raise Exception("Source '%s' not found" % source)

    def __compare_files(self, filePathA, filePathB):
        """Compares two files, ignoring end of lines in text files.

        :param filePathA: Path to first file.
        :param filePathB: Path to second file.
        :return: True if files are equal.
        """
        try:
            with open(filePathA, "r") as file1, open(filePathB, "r") as file2:
                while True:
                    line1 = next(file1, None)
                    line2 = next(file2, None)
                    # if either file is finished return true
                    if line1 is None or line2 is None:
                        return line1 is None and line2 is None
                    # compare lines regardless of EOL
                    if line1.rstrip("\r\n") != line2.rstrip("\r\n"):
                        return False

        # do binary comparison if there are invalid characters
        except UnicodeDecodeError:
            return filecmp.cmp(filePathA, filePathB, shallow=False)

        except FileNotFoundError:
            return False

    def __compare_objects(self, objectA, objectB):
        """Compares two files or two directories.

        :param objectA: Path to first file or directory.
        :param objectB: Path to second file or directory.
        :return: True if objects are equal.
        """
        if os.path.isfile(objectA) and os.path.isfile(objectB):
            return self.__compare_files(objectA, objectB)

        objectA = os.path.relpath(objectA)
        all_files = self.get_files(objectA)

        for filepath in all_files:
            destPath = os.path.join(objectB, filepath[len(objectA) + 1 :])
            if not self.__compare_files(filepath, destPath):
                return False

        return True

    @staticmethod
    def __link_object(target, link, actual_target):
        """Creates symbolic link to a file or directory.

        :param target: Path to target file or directory.
        :param link: Path to symbolic link.
        :param actual_target: Path to actual target file or directory.
        """
        isdir = os.path.isdir(actual_target)
        try:
            os.symlink(target, link, target_is_directory=isdir)

        except OSError as e:
            log.error(
                "Failed to create %s symoblic link '%s => %s': %s"
                % ("directory" if isdir else "file", link, target, str(e))
            )
            return None
        return isdir

    @staticmethod
    def __replace_vars(pathstr):
        """Replaces tokens in string with environment variable or config default.

        :param pathstr: Path string with tokens.
        :return: Path string with tokens replaced.
        """
        while True:
            openBracket = pathstr.find(config.PATH_TOKEN_OPEN)
            closeBracket = pathstr.find(config.PATH_TOKEN_CLOSE)
            if -1 == openBracket or closeBracket <= openBracket:
                return pathstr
            var = pathstr[openBracket + 1 : closeBracket].upper()
            replacement = os.getenv(var, config.DEFAULT_ENV.get(var))
            if not replacement:
                raise Exception("Cannot resolve env var '%s'" % var)
            pathstr = pathstr[0:openBracket] + replacement + pathstr[closeBracket + 1 :]

    def get_files(self, start):
        """Returns the list of files to be disted.

        :param start: Starting directory.
        :return: List of relative file paths.
        """
        all_files = [f for f in util.walk(start)]
        repo_files = []

        if self.repo:
            try:
                repo_files = [f for f in self.get_repo_files(start)]
                tracked_dirs = [os.path.dirname(f) for f in repo_files]
                untracked_files = [f for f in all_files if f not in repo_files]
                untracked_dirs = [
                    os.path.dirname(f)
                    for f in untracked_files
                    if os.path.dirname(f) not in tracked_dirs
                ]
                untracked_files = [
                    f
                    for f in untracked_files
                    if os.path.dirname(f) not in untracked_dirs
                ]

                if untracked_files or untracked_dirs:
                    log.warning("Untracked files:")
                    for f in untracked_files:
                        log.warning("  %s" % f)
                    for d in untracked_dirs:
                        log.warning("  %s/" % d)

            except Exception as e:
                log.warning("Failed to get files from repo: %s" % str(e))

        return all_files

    def dist(
        self,
        target=None,
        show=False,
        force=False,
        yes=False,
        dryrun=False,
        verbose=False,
    ):
        """Performs the file distribution.

        :param target: optionally match specific targets.
        :param show: Show file versions.
        :param force: Force distribution.
        :param yes: Assume yes to all questions.
        :param dryrun: Perform dry run.
        :param verbose: Show more information.
        :return: True if successful.
        """
        if self.root is None:
            log.error("dist file not found or invalid")
            return False

        if not self.read_git_info():
            return False

        targets_node = self.get_targets()

        if targets_node is None:
            return False

        changed_files = self.git_changed_files()
        if config.DIST_FILE in changed_files:
            log.info(
                "WARNING: %s should be checked into git repository!" % config.DIST_FILE
            )

        targets = []
        for target_name, target_dict in targets_node.items():
            source = target_dict.get(config.TAG_SOURCEPATH)
            dest = target_dict.get(config.TAG_DESTPATH)
            if source is None or dest is None:
                log.info(
                    "Target %s: Missing '%s' or '%s' tag"
                    % (target_name, config.TAG_SOURCEPATH, config.TAG_DESTPATH)
                )
                self.root = None
                return False

            # optionally match on specific targets
            if target and not fnmatch.fnmatch(target_name, target):
                continue

            try:
                dest = util.normalize_path(self.__replace_vars(dest))
            except Exception as e:
                log.info(
                    "%s in <%s> for %s" % (str(e), config.TAG_DESTPATH, target_name)
                )
                return False

            # make sure file exists
            if not os.path.exists(self.directory + os.path.sep + source):
                log.info(
                    "Target %s: Source '%s' does not exist" % (target_name, source)
                )
                return False

            # check if file has uncommitted changes
            if (
                not show
                and not force
                and os.path.abspath(self.directory + os.path.sep + source)
            ) in changed_files:
                log.info(
                    "Target %s: Source '%s' has uncommitted changes.  "
                    "Commit the changes or use --force." % (target_name, source)
                )
                return False

            # create destination directory if it does not exist (or exit)
            dest_dir = os.path.dirname(dest)
            if not show and not dryrun and not os.path.exists(dest_dir):
                question = (
                    "Target %s: Destination directory '%s' does not "
                    "exist, create it now?" % (target_name, dest_dir)
                )
                if not yes and not util.yesNo(question):
                    return False
                try:
                    os.makedirs(dest_dir)
                except Exception as e:
                    log.info(
                        "ERROR: Failed to create directory '%s': %s"
                        % (dest_dir, str(e))
                    )
                    return False
            targets.append((source, dest))

        if not targets:
            if target:
                log.info("Target %s not found in %s", target, config.DIST_FILE)
            else:
                log.info("No targets found in %s", config.DIST_FILE)
            self.root = False
            return False

        # check if local git repo is behind remote
        if not show and not force and self.is_git_behind():
            return False

        if dryrun:
            log.info(config.DRYRUN_MESSAGE)

        # process targets listed in dist file
        for source, dest in targets:
            util.create_dest_folder(dest, dryrun, yes)

            # write the dist info file
            if not dryrun and not show:
                info = {
                    "name": self.name,
                    "origin": self.path,
                    "source": source,
                    "author": self.author,
                }
                util.write_dist_info(dest, info)

            # define dist version information
            version_num = 0
            version_file = ""
            version_list = self.__get_file_versions(dest)

            # TODO: make show a separate method
            if show:
                if callable(getattr(os, "readlink", None)):
                    if not os.path.lexists(dest):
                        log.info("Missing: %s" % dest)
                    else:
                        log.info("%s => %s:" % (source, os.readlink(dest)))
                else:
                    log.info("%s:" % source)

                for version_file, version_num, version_commit in version_list:
                    log.info(
                        "%s: %s - %s"
                        % (
                            version_num,
                            version_file,
                            time.ctime(os.path.getmtime(version_file)),
                        )
                    )
                    if self.repo and verbose:
                        try:
                            commit = self.repo.commit(version_commit)
                            log.info("    %s" % commit.message.strip())
                            log.info(
                                "    %s - %s"
                                % (time.ctime(commit.committed_date), commit.author)
                            )
                        except Exception:
                            pass
                continue

            # relative path to the source file
            if source == ".":
                source_path = self.directory
            else:
                source_path = os.path.join(self.directory, source)

            if version_list:
                version_file, version_num, _ = version_list[-1]
                if version_file and self.__compare_objects(source_path, version_file):
                    if os.path.exists(dest) and os.path.lexists(dest):
                        log.info("Unchanged: %s => %s" % (source, version_file))
                    else:
                        question = (
                            "Target %s: link '%s' missing or broken,"
                            " fix it now?" % (target_name, dest)
                        )
                        if yes or util.yesNo(question):
                            if dryrun:
                                log.info("Fixed: %s => %s" % (source, version_file))
                            else:
                                if os.path.lexists(dest):
                                    util.remove_object(dest)
                                isdir = self.__link_object(
                                    config.DIR_VERSIONS
                                    + os.path.sep
                                    + os.path.basename(version_file),
                                    dest,
                                    version_file,
                                )
                                if isdir is not None:
                                    log.info(
                                        "Fixed: %s =%s> %s"
                                        % (
                                            source,
                                            "d" if isdir else "f",
                                            version_file,
                                        )
                                    )
                    continue

                version_num += 1

            # copy source file to the versioned destination
            versions_dir = os.path.dirname(dest) + "/" + config.DIR_VERSIONS
            if not dryrun and not os.path.exists(versions_dir):
                os.mkdir(versions_dir)
            version_dest = (
                versions_dir + "/" + os.path.basename(dest) + "." + str(version_num)
            )
            if self.short_head:
                version_dest += "." + self.short_head
                # note in file name if forced (not synced with current head)
                # if force:
                #     version_dest += "-forced"
            # note in file name if not on a main/master branch
            # FIXME: breaks if there is a / in the branch name (replace special chars)
            # if self.branch_name and self.branch_name not in config.MAIN_BRANCHES:
            #     version_dest += "." + self.branch_name
            # copy the file/directory to the versioned location
            if not dryrun:
                self.__copy_object(source_path, version_dest)
            # delete existing symbolic link if it exists
            if not dryrun and os.path.lexists(dest):
                util.remove_object(dest)
            # create the new symbolic link
            if dryrun:
                log.info("Updated: %s => %s" % (source, version_dest))
            else:
                isdir = self.__link_object(
                    config.DIR_VERSIONS + os.path.sep + os.path.basename(version_dest),
                    dest,
                    version_dest,
                )
                if isdir is not None:
                    log.info(
                        "Updated: %s =%s> %s"
                        % (source, "d" if isdir else "f", version_dest)
                    )

        if self.repo:
            self.repo.close()

        return True

    def reset_file_version(self, target, dryrun=False):
        """Resets the symbolic link of a versioned file to point to the latest.

        :param target: target name in dist file.
        :param dryrun: Perform dry run.
        :return: True on success, False on failure.
        """
        targets_node = self.get_targets()

        if targets_node is None:
            return False

        if dryrun:
            log.info(config.DRYRUN_MESSAGE)

        any_found = False
        for target_name, target_dict in targets_node.items():
            if (
                target_dict.get(config.TAG_SOURCEPATH) is None
                or target_dict.get(config.TAG_DESTPATH) is None
            ):
                continue
            source = util.normalize_path(target_dict.get(config.TAG_SOURCEPATH))
            try:
                dest = util.normalize_path(
                    self.__replace_vars(target_dict.get(config.TAG_DESTPATH))
                )
            except Exception as e:
                log.info(
                    "%s in <%s> for %s" % (str(e), config.TAG_DESTPATH, target_name)
                )
                return False

            if target and target != target_name:
                continue

            any_found = True
            version_list = self.__get_file_versions(dest)
            if not version_list:
                log.info(
                    "Target %s: No versioned files found for '%s'"
                    % (target_name, source)
                )
            else:
                verfile = version_list[-1][0]
                # remove existing symbolic link
                if not dryrun and os.path.lexists(dest):
                    util.remove_object(dest)
                # create new link to point to requested file
                if dryrun:
                    log.info("%s => %s" % (source, verfile))
                else:
                    isdir = self.__link_object(
                        config.DIR_VERSIONS + os.path.sep + os.path.basename(verfile),
                        dest,
                        verfile,
                    )
                    if isdir is not None:
                        log.info(
                            "%s =%s> %s" % (source, "d" if isdir else "f", verfile)
                        )

        if not any_found:
            log.info("No targets found to reset")

        return any_found

    def change_file_version(self, target, target_commit, target_version, dryrun=False):
        """Changes the symbolic link of a versioned file to point to a different file.

        :param target: target name in dist file.
        :param target_commit: The commit hash to point to.
        :param target_version: The version number to point to.
        :param dryrun: Perform dry run.
        :return: True on success, False on failure.
        """
        targets_node = self.get_targets()

        if targets_node is None:
            return False

        if dryrun:
            log.info(config.DRYRUN_MESSAGE)

        any_found = False
        for target_name, target_dict in targets_node.items():
            if (
                target_dict.get(config.TAG_SOURCEPATH) is None
                or target_dict.get(config.TAG_DESTPATH) is None
            ):
                continue

            source = util.normalize_path(target_dict.get(config.TAG_SOURCEPATH))
            try:
                dest = util.normalize_path(
                    self.__replace_vars(target_dict.get(config.TAG_DESTPATH))
                )
            except Exception as e:
                log.error(
                    "%s in <%s> for %s" % (str(e), config.TAG_DESTPATH, target_name)
                )
                return False

            if target and target != target_name:
                continue

            version_list = self.__get_file_versions(dest)
            if not version_list:
                log.info(
                    "Target %s: No versioned files found for '%s'"
                    % (target_name, source)
                )
                continue
            any_found_this_target = False

            if target_version < 0:
                if abs(target_version) > len(version_list) - 1:
                    log.warning(
                        "Requested to roll back %d versions but there "
                        "are only %d previous versions for %s"
                        % (abs(target_version), len(version_list) - 1, source)
                    )
                    continue
                target_version = version_list[target_version - 1][1]

            for verfile, vernum, vercommit in version_list:
                if (not target_commit and vernum == target_version) or (
                    target_commit and self.__hashes_equal(target_commit, vercommit)
                ):
                    # found matching versioned target
                    any_found = True
                    any_found_this_target = True
                    # remove existing symbolic link
                    if not dryrun and os.path.lexists(dest):
                        util.remove_object(dest)
                    # create new symbolic link to point to requested versioned file
                    if dryrun:
                        log.info("%s => %s" % (source, verfile))
                    else:
                        isdir = self.__link_object(
                            config.DIR_VERSIONS
                            + os.path.sep
                            + os.path.basename(verfile),
                            dest,
                            verfile,
                        )
                        if isdir is not None:
                            log.info(
                                "%s =%s> %s" % (source, "d" if isdir else "f", verfile)
                            )
                    break

            if not any_found_this_target:
                if target_commit:
                    log.info(
                        "Target commit %s not found for target %s"
                        % (target_commit, target_name)
                    )
                else:
                    log.info(
                        "Target version %d not found for target %s"
                        % (target_version, target_name)
                    )

        if not any_found:
            log.info("No targets found to change version")

        return any_found

    def delete_target(self, target, yes=False, dryrun=False):
        """Delete a target's destination files. Deletes symlink, .dist and version
        files/directories.

        :param target: target name in dist file.
        :param yes: Answer yes to all questions.
        :param dryrun: Perform dry run.
        :return: True on success, False on failure.
        """
        targets_node = self.get_targets()

        if targets_node is None:
            return False

        if dryrun:
            log.info(config.DRYRUN_MESSAGE)

        any_found = False
        for target_name, target_dict in targets_node.items():
            if (
                target_dict.get(config.TAG_SOURCEPATH) is not None
                and target_dict.get(config.TAG_DESTPATH) is not None
            ):
                if target and target != target_name:
                    continue

                source = util.normalize_path(target_dict.get(config.TAG_SOURCEPATH))
                try:
                    dest = util.normalize_path(
                        self.__replace_vars(target_dict.get(config.TAG_DESTPATH))
                    )
                except Exception as e:
                    log.info(
                        "%s in <%s> for %s" % (str(e), config.TAG_DESTPATH, target_name)
                    )
                    return False
                version_list = self.__get_file_versions(dest)
                question = "Delete target '%s' (%s => %s) and %d versions?" % (
                    target_name,
                    source,
                    dest,
                    len(version_list),
                )
                if yes or dryrun or util.yesNo(question):
                    any_found = True
                    distinfo = util.get_dist_info(dest=dest)
                    if os.path.lexists(dest):
                        log.info("Deleting: %s" % dest)
                        if not dryrun:
                            util.remove_object(dest)
                    else:
                        log.info("Missing: %s" % dest)
                    if os.path.lexists(distinfo):
                        log.info("Deleting: %s" % distinfo)
                        if not dryrun:
                            os.remove(distinfo)
                    else:
                        log.info("Missing: %s" % distinfo)
                    for verFile, _, _ in version_list:
                        log.info("Deleting: %s" % verFile)
                        if not dryrun:
                            util.remove_object(verFile, recurse=True)

        if not any_found:
            log.info("No targets found to delete")

        return any_found



#!/usr/bin/env python
#
# Copyright (c) 2024, Ryan Galloway (ryan@rsgalloway.com)
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#  - Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
#  - Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
#  - Neither the name of the software nor the names of its contributors
#    may be used to endorse or promote products derived from this software
#    without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#

__doc__ = """
Contains tests for the util module.
"""

import os
import unittest
from unittest.mock import patch

from distman.util import (check_symlinks, full_path, get_user, is_file_hidden,
                          is_ignorable, normalize_path, remove_object, walk,
                          yesNo)


class TestUtils(unittest.TestCase):
    def test_check_symlinks(self):
        # Mock the tempfile.mktemp() function to return temporary file paths
        with patch("tempfile.mktemp") as mock_mktemp:
            mock_mktemp.side_effect = ["/tmp/temp_file", "/tmp/link_file"]

            # Mock the os.symlink() function to raise an OSError
            with patch("os.symlink", side_effect=OSError):
                result = check_symlinks()

                # Assert that the method returns False
                self.assertFalse(result)

            # Assert that the temporary files are removed
            self.assertFalse(os.path.exists("/tmp/temp_file"))
            self.assertFalse(os.path.exists("/tmp/link_file"))

    def test_get_user(self):
        # Mock the os.getenv() function to return the user name
        with patch("os.getenv") as mock_getenv:
            mock_getenv.side_effect = ["john", None, None]

            # Call the method to get the user name
            result = get_user()

            # Assert that the correct user name is returned
            self.assertEqual(result, "john")

            # Call the method when both USER and USERNAME environment variables are not set
            result = get_user()

            # Assert that the default user name is returned
            self.assertEqual(result, "unknown")

    def test_is_file_hidden(self):
        # Mock the os.path.basename() and has_hidden_attr() functions
        with patch("os.path.basename") as mock_basename, patch(
            "distman.util.has_hidden_attr"
        ) as mock_has_hidden_attr:
            # Set the return values of the mocked functions
            mock_basename.return_value = ".hidden_file"
            mock_has_hidden_attr.return_value = True

            # Call the method to check if the file is hidden
            result = is_file_hidden("/path/to/file")

            # Assert that the method returns True
            self.assertTrue(result)

    def test_is_ignorable(self):
        # Mock the is_file_hidden() function
        with patch("distman.util.is_file_hidden") as mock_is_file_hidden:
            # Set the return value of the mocked function
            mock_is_file_hidden.return_value = False

            # Call the method to check if the file is ignorable
            result = is_ignorable("/path/to/file")

            # Assert that the method returns False
            self.assertFalse(result)

    def test_normalize_path(self):
        # Call the method to normalize a path
        result = normalize_path("path\\to\\file/")

        # Assert that the path is normalized correctly
        self.assertEqual(result, "path/to/file")

    def test_full_path(self):
        # Call the method to get the full path from a relative path
        result = full_path("/start/directory", "../relative/path")

        # Assert that the full path is returned correctly
        self.assertEqual(result, "/start/relative/path")

    def test_remove_object(self):
        # Mock the os.path.isdir() and os.remove() functions
        with patch("os.path.isdir") as mock_isdir, patch("os.remove") as mock_remove:
            # Set the return value of the mocked isdir() function
            mock_isdir.return_value = True

            # Call the method to remove a directory
            remove_object("/path/to/directory", recurse=True)

            # Assert that the rmtree() function is called
            mock_remove.assert_called_with("/path/to/directory")

            # Call the method to remove a file
            remove_object("/path/to/file")

            # Assert that the remove() function is called
            mock_remove.assert_called_with("/path/to/file")

    def test_yesNo(self):
        # Mock the input() function to return user input
        with patch("builtins.input") as mock_input:
            # Set the return value of the mocked input() function
            mock_input.return_value = "y"

            # Call the method to ask a yes/no question
            result = yesNo("Do you want to continue?")

            # Assert that the method returns True
            self.assertTrue(result)

            # Set the return value of the mocked input() function
            mock_input.return_value = "n"

            # Call the method to ask a yes/no question
            result = yesNo("Do you want to continue?")

            # Assert that the method returns False
            self.assertFalse(result)

    def test_walk(self):
        # Mock the is_ignorable() function
        with patch("distman.util.is_ignorable") as mock_is_ignorable:
            # Set the return value of the mocked is_ignorable() function
            mock_is_ignorable.return_value = False

            # Call the method to walk through the files
            result = list(walk("/path/to/directory"))

            # Assert that the correct file paths are returned
            self.assertEqual(
                result,
                [
                    "/path/to/directory/file1.txt",
                    "/path/to/directory/file2.txt",
                    "/path/to/directory/subdirectory/file3.txt",
                ],
            )


if __name__ == "__main__":
    unittest.main()
